/* fplll/fplll_config.h.  Generated from fplll_config.h.in by configure.  */
#ifndef FPLLL_CONFIG__H
#define FPLLL_CONFIG__H

/* use quaddouble library */
#define FPLLL_WITH_QD 1

/* fplll major version */
#define FPLLL_MAJOR_VERSION 5

/* fplll minor version */
#define FPLLL_MINOR_VERSION 5

/* fplll micro version */
#define FPLLL_MICRO_VERSION 0

/* fplll version */
#define FPLLL_VERSION 5.5.0

/* fplll version info */
#define FPLLL_VERSION_INFO 

/* Maximum supported enumeration dimension */
#define FPLLL_MAX_ENUM_DIM 256

/* Recursive enumeration enabled */
#define FPLLL_WITH_RECURSIVE_ENUM 1

/* Maximum supported parallel enumeration dimension */
#define FPLLL_MAX_PARALLEL_ENUM_DIM 120

/* Switch between GMP and MPIR */
#define HAVE_LIBGMP 1
/* #undef HAVE_LIBMPIR */

/* external enumeration */
/* #undef FPLLL_EXTENUM_FUNC */

#endif //FPLLL_CONFIG__H
